unit imDBImage;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, DB,
  DBCtrls, JPEG;

type
  TimDBImage = class(TCustomControl)
  private
    FAutoDisplay: Boolean;
    FBorderStyle: TBorderStyle;
    FCenter: Boolean;
    FQuality: TJPEGQualityRange;
    FDataLink: TFieldDataLink;
    FPicture: TPicture;
    FPictureLoaded: Boolean;
    FQuickDraw: Boolean;
    FStretch: Boolean;
    procedure DataChange(Sender: TObject);
    function GetDataField: string;
    function GetDataSource: TDataSource;
    function GetField: TField;
    function GetReadOnly: Boolean;
    procedure PictureChanged(Sender: TObject);
    procedure SetAutoDisplay(Value: Boolean);
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetCenter(Value: Boolean);
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    procedure SetPicture(Value: TPicture);
    procedure SetReadOnly(Value: Boolean);
    procedure SetStretch(Value: Boolean);
    procedure UpdateData(Sender: TObject);
    procedure CMEnter(var M: TCMEnter); message CM_ENTER;
    procedure CMExit(var M: TCMExit); message CM_EXIT;
    procedure CMTextChanged(var M: TMessage); message CM_TEXTCHANGED;
    procedure WMLButtonDown(var M: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonDblClk(var M: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
    procedure WMCut(var M: TMessage); message WM_CUT;
    procedure WMCopy(var M: TMessage); message WM_COPY;
    procedure WMPaste(var M: TMessage); message WM_PASTE;
    procedure WMSize(var M: TMessage); message WM_SIZE;
  protected
    procedure AssignFieldToPicture(APicture: TPicture);
    procedure AssignPictureToField;
    procedure CreateParams(var Params: TCreateParams); override;
    function GetPalette: HPALETTE; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure CopyToClipboard;
    procedure CutToClipboard;
    procedure LoadPicture;
    procedure PasteFromClipboard;
    property Field: TField read GetField;
    property Picture: TPicture read FPicture write SetPicture;
  published
    property Align;
    property AutoDisplay: Boolean read FAutoDisplay write SetAutoDisplay default True;
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsSingle;
    property Center: Boolean read FCenter write SetCenter default True;
    property Color;
    property CompressionQuality: TJPEGQualityRange read FQuality write FQuality default 90;
    property Ctl3D;
    property DataField: string read GetDataField write SetDataField;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property ParentColor default False;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly default False;
    property QuickDraw: Boolean read FQuickDraw write FQuickDraw default True;
    property ShowHint;
    property Stretch: Boolean read FStretch write SetStretch default False;
    property TabOrder;
    property TabStop default True;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

procedure Register;

implementation

uses Clipbrd;

procedure Register;
begin
  RegisterComponents('Data Controls', [TimDBImage]);
end;

constructor TimDBImage.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque, csReplicatable];
  if not NewStyleControls then ControlStyle := ControlStyle + [csFramed];
  Width := 105;
  Height := 105;
  TabStop := True;
  ParentColor := False;
  FPicture := TPicture.Create;
  FPicture.OnChange := PictureChanged;
  FBorderStyle := bsSingle;
  FAutoDisplay := True;
  FCenter := True;
  FDataLink := TFieldDataLink.Create;
  FDataLink.Control := Self;
  FDataLink.OnDataChange := DataChange;
  FDataLink.OnUpdateData := UpdateData;
  FQuickDraw := True;
  FQuality := 90;
end;

destructor TimDBImage.Destroy;
begin
  FPicture.Free;
  FDataLink.Free;
  FDataLink := nil;
  inherited Destroy;
end;

procedure TimDBImage.AssignFieldToPicture(APicture: TPicture);
type
  TGraphicHeader = record
    Count: Word;                { Fixed at 1 }
    HType: Word;                { Fixed at $0100 }
    Size: Longint;              { Size not including header }
  end;
var
  BS: TStream;

  procedure LoadAsJPEG;
  begin
    BS.Position := 0;
    APicture.Graphic := TJPEGImage.Create;
    try
      APicture.Graphic.LoadFromStream(BS);
    except
      APicture.Graphic := nil;
      raise;
    end;
  end;

  procedure LoadAsBMP;
  begin
    try
      APicture.Graphic := TBitmap.Create;
      APicture.Graphic.LoadFromStream(BS);
    except
      LoadAsJPEG;
    end;
  end;

var
  Size: LongInt;
  Header: TGraphicHeader;
  BmpH: TBitmapFileHeader;
  IsBitmap: Boolean;
begin
  if FDataLink.Field.IsNull then
  begin
    APicture.Graphic := nil;
    Exit;
  end;
  BS := FDataLink.DataSet.CreateBlobStream(TBlobField(FDataLink.Field), bmRead);
  try
    Size := BS.Size;
    if Size >= SizeOf(TGraphicHeader) then
    begin
      BS.Read(Header, SizeOf(Header));
      if (Header.Count <> 1) or (Header.HType <> $0100) or
        (Header.Size <> Size - SizeOf(Header)) then
        BS.Position := 0;
    end;
    IsBitmap := True;
    if Size - BS.Position < SizeOf(TBitmapFileHeader) then
      IsBitmap := False
    else
    begin
      BS.Read(BmpH, SizeOf(BmpH));
      if BmpH.bfType <> $4D42 then IsBitmap := False;
    end;
    if IsBitmap then
    begin
      BS.Position := BS.Position - SizeOf(BmpH);
      LoadAsBmp;
    end
    else
      LoadAsJPEG;
  finally
    BS.Free;
  end;
end;

procedure TimDBImage.AssignPictureToField;
var
  BS: TStream;
begin
  BS := FDataLink.DataSet.CreateBlobStream(TBlobField(FDataLink.Field), bmWrite);
  try
    TJPEGImage(Picture.Graphic).CompressionQuality := FQuality;
    Picture.Graphic.SaveToStream(BS);
  finally
    BS.Free;
  end;
end;

procedure TimDBImage.CMEnter(var M: TCMEnter);
begin
  Invalidate; { Draw the focus marker }
  inherited;
end;

procedure TimDBImage.CMExit(var M: TCMExit);
begin
  try
    FDataLink.UpdateRecord;
  except
    SetFocus;
    raise;
  end;
  Invalidate; { Erase the focus marker }
  inherited;
end;

procedure TimDBImage.CMTextChanged(var M: TMessage);
begin
  inherited;
  if not FPictureLoaded then Invalidate;
end;

procedure TimDBImage.CopyToClipboard;
begin
  if Picture.Graphic <> nil then Clipboard.Assign(Picture);
end;

procedure TimDBImage.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    if FBorderStyle = bsSingle then
      if NewStyleControls and Ctl3D then
        ExStyle := ExStyle or WS_EX_CLIENTEDGE
      else
        Style := Style or WS_BORDER;
    WindowClass.Style := WindowClass.Style and not (CS_HREDRAW or CS_VREDRAW);
  end;
end;

procedure TimDBImage.CutToClipboard;
begin
  if Picture.Graphic <> nil then
    if FDataLink.Edit then
    begin
      CopyToClipboard;
      Picture.Graphic := nil;
    end;
end;

procedure TimDBImage.DataChange(Sender: TObject);
begin
  Picture.Graphic := nil;
  FPictureLoaded := False;
  if FAutoDisplay then LoadPicture;
end;

function TimDBImage.GetDataField: string;
begin
  Result := FDataLink.FieldName;
end;

function TimDBImage.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

function TimDBImage.GetField: TField;
begin
  Result := FDataLink.Field;
end;

function TimDBImage.GetPalette: HPALETTE;
begin
  Result := 0;
  if FPicture.Graphic is TBitmap then
    Result := TBitmap(FPicture.Graphic).Palette
  else if FPicture.Graphic is TJPEGImage then
    Result := TJPEGImage(FPicture.Graphic).Palette;
end;

function TimDBImage.GetReadOnly: Boolean;
begin
  Result := FDataLink.ReadOnly;
end;

procedure TimDBImage.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
  case Key of
    VK_INSERT:
      if ssShift in Shift then
        PasteFromClipBoard
      else if ssCtrl in Shift then
        CopyToClipBoard;
    VK_DELETE:
      if ssShift in Shift then CutToClipBoard;
  end;
end;

procedure TimDBImage.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);
  case Key of
    ^X: CutToClipBoard;
    ^C: CopyToClipBoard;
    ^V: PasteFromClipBoard;
    #13: LoadPicture;
    #27: FDataLink.Reset;
  end;
end;

procedure TimDBImage.LoadPicture;
begin
  if not FPictureLoaded and (not Assigned(FDataLink.Field) or
    FDataLink.Field.IsBlob) then
    AssignFieldToPicture(Picture);
end;

procedure TimDBImage.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (FDataLink <> nil) and
    (AComponent = DataSource) then DataSource := nil;
end;

procedure TimDBImage.Paint;
var
  Size: TSize;
  R: TRect;
  S: string;
  DrawPict: TPicture;
  Form: TCustomForm;
  Pal: HPalette;
begin
  with Canvas do
  begin
    Brush.Style := bsSolid;
    Brush.Color := Color;
    if FPictureLoaded or (csPaintCopy in ControlState) then
    begin
      DrawPict := TPicture.Create;
      Pal := 0;
      try
        if (csPaintCopy in ControlState) and
          Assigned(FDataLink.Field) and FDataLink.Field.IsBlob then
        begin
          AssignFieldToPicture(DrawPict);
          if DrawPict.Graphic is TBitmap then
            DrawPict.Bitmap.IgnorePalette := QuickDraw;
        end
        else
        begin
          DrawPict.Assign(Picture);
          if Focused and (DrawPict.Graphic <> nil) and (DrawPict.Graphic.Palette <> 0) then
          begin { Control has focus, so realize the bitmap palette in foreground }
            Pal := SelectPalette(Handle, DrawPict.Graphic.Palette, False);
            RealizePalette(Handle);
          end;
        end;
        if Stretch then
          if (DrawPict.Graphic = nil) or DrawPict.Graphic.Empty then
            FillRect(ClientRect)
          else
            StretchDraw(ClientRect, DrawPict.Graphic)
        else
        begin
          SetRect(R, 0, 0, DrawPict.Width, DrawPict.Height);
          if Center then OffsetRect(R, (ClientWidth - DrawPict.Width) div 2,
            (ClientHeight - DrawPict.Height) div 2);
          StretchDraw(R, DrawPict.Graphic);
          ExcludeClipRect(Handle, R.Left, R.Top, R.Right, R.Bottom);
          FillRect(ClientRect);
          SelectClipRgn(Handle, 0);
        end;
      finally
        if Pal <> 0 then SelectPalette(Handle, Pal, True);
        DrawPict.Free;
      end;
    end
    else begin
      Font := Self.Font;
      if FDataLink.Field <> nil then
        S := FDataLink.Field.DisplayLabel
      else
        S := Name;
      S := '(' + S + ')';
      Size := TextExtent(S);
      R := ClientRect;
      TextRect(R, (R.Right - Size.cx) div 2, (R.Bottom - Size.cy) div 2, S);
    end;
    Form := GetParentForm(Self);
    if (Form <> nil) and (Form.ActiveControl = Self) and
      not (csDesigning in ComponentState) and
      not (csPaintCopy in ControlState) then
    begin
      Brush.Color := clWindowFrame;
      FrameRect(ClientRect);
    end;
  end;
end;

procedure TimDBImage.PasteFromClipboard;
var
  B: TBitmap;
begin
  if Clipboard.HasFormat(CF_BITMAP) and FDataLink.Edit then
  begin
    B := TBitmap.Create;
    try
      B.Assign(Clipboard);
      Picture.Graphic := TJPEGImage.Create;
      Picture.Graphic.Assign(B);
      TJPEGImage(Picture.Graphic).Compress;
    finally
      B.Free;
    end;
  end;
end;

procedure TimDBImage.PictureChanged(Sender: TObject);
begin
  if FPictureLoaded then FDataLink.Modified;
  FPictureLoaded := True;
  Invalidate;
end;

procedure TimDBImage.SetAutoDisplay(Value: Boolean);
begin
  if FAutoDisplay <> Value then
  begin
    FAutoDisplay := Value;
    if Value then LoadPicture;
  end;
end;

procedure TimDBImage.SetBorderStyle(Value: TBorderStyle);
begin
  if FBorderStyle <> Value then
  begin
    FBorderStyle := Value;
    RecreateWnd;
  end;
end;

procedure TimDBImage.SetCenter(Value: Boolean);
begin
  if FCenter <> Value then
  begin
    FCenter := Value;
    Invalidate;
  end;
end;

procedure TimDBImage.SetDataField(const Value: string);
begin
  FDataLink.FieldName := Value;
end;

procedure TimDBImage.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource := Value;
  if Value <> nil then Value.FreeNotification(Self);
end;

procedure TimDBImage.SetPicture(Value: TPicture);
begin
  FPicture.Assign(Value);
end;

procedure TimDBImage.SetReadOnly(Value: Boolean);
begin
  FDataLink.ReadOnly := Value;
end;

procedure TimDBImage.SetStretch(Value: Boolean);
begin
  if FStretch <> Value then
  begin
    FStretch := Value;
    Invalidate;
  end;
end;

procedure TimDBImage.UpdateData(Sender: TObject);
begin
  if Picture.Graphic is TBitmap then
    FDataLink.Field.Assign(Picture.Graphic)
  else if Picture.Graphic is TJPEGImage then
    AssignPictureToField
  else
    FDataLink.Field.Clear;
end;

procedure TimDBImage.WMLButtonDown(var M: TWMLButtonDown);
begin
  if TabStop and CanFocus then SetFocus;
  inherited;
end;

procedure TimDBImage.WMLButtonDblClk(var M: TWMLButtonDblClk);
begin
  LoadPicture;
  inherited;
end;

procedure TimDBImage.WMCut(var M: TMessage);
begin
  CutToClipboard;
end;

procedure TimDBImage.WMCopy(var M: TMessage);
begin
  CopyToClipboard;
end;

procedure TimDBImage.WMPaste(var M: TMessage);
begin
  PasteFromClipboard;
end;

procedure TimDBImage.WMSize(var M: TMessage);
begin
  inherited;
  Invalidate;
end;

end.
